/**
 * Lecture notes:
 * - SDOT example.
 * - Illustrates the effect of warp divergence.
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 512
#define NUM_BLOCKS 1024



/// dot product of two vectors (CPU VERSION)
float sdot_ref(int num_elements, float * const  x, float* const y);


/// compares the results of the GPU version against CPU
void check_result(int num_elements, float * const x, float * const y, float result);



/// dot product of two vectors (GPU VERSION)
/// With warp divergence
__global__ void sdot_bad(int num_elements, float * const  x, float*const y, float * result)
{
   int idx = blockIdx.x * blockDim.x + threadIdx.x;

   __shared__ float reduce[BLOCKSIZE];
   reduce[threadIdx.x] = y[idx] * x[idx]; //store local result to memory

   int offset = 2;

   while(offset <= BLOCKSIZE)
   {
      __syncthreads();
      if(threadIdx.x % offset == 0)
         reduce[threadIdx.x] += reduce[threadIdx.x + offset/2];

      offset *= 2;
   }

   //store result to global memory
   if(threadIdx.x == 0)
      result[blockIdx.x] = reduce[0];
}






/// dot product of two vectors (GPU VERSION)
/// NO warp divergence
__global__ void sdot_good(int num_elements, float * const  x, float*const y, float * result)
{
   int idx = blockIdx.x * blockDim.x + threadIdx.x;

   __shared__ float reduce[BLOCKSIZE];
   reduce[threadIdx.x] = y[idx] * x[idx]; //store local result to memory

   int stride = BLOCKSIZE/2;

   while(stride > 0)
   {
      __syncthreads();
      if(threadIdx.x < stride)
         reduce[threadIdx.x] += reduce[threadIdx.x + stride];

      stride /= 2;
   }

   //store result to global memory
   if(threadIdx.x == 0)
      result[blockIdx.x] = reduce[0];
}




int main(int argc, char** argv)
{
   int i;
   int num_elements = NUM_BLOCKS * BLOCKSIZE;

   int size = sizeof(float) * num_elements;
   printf("Allocates %f MB\n", 2.0f * ((float)size)/1024/1024);


   //allocate host memory
   float *x_h, *y_h, *buffer_h;
   x_h = (float*) malloc(size);
   y_h = (float*) malloc(size);
   buffer_h = (float*) malloc(sizeof(float) * NUM_BLOCKS);


   //init x and y with random numbers in [0,1]
   for(i=0;i < num_elements; ++i){
      x_h[i] = ((float)rand())/RAND_MAX;
      y_h[i] = ((float)rand())/RAND_MAX;
   }


   //allocate device memory
   float *x_d, *y_d, *buffer_d;
   HANDLE_ERROR( cudaMalloc(&x_d, size) );
   HANDLE_ERROR( cudaMalloc(&y_d, size) );
   HANDLE_ERROR( cudaMalloc(&buffer_d, sizeof(float) * NUM_BLOCKS) );


   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(x_d, x_h, size, cudaMemcpyHostToDevice) ); 
   HANDLE_ERROR( cudaMemcpy(y_d, y_h, size, cudaMemcpyHostToDevice) ); 


   float min_time = FLT_MAX;
   int nIter = 20;
   //take minimum over nIter runs
   for(int i=0;i < nIter; ++i){

      float time;
      cudaEvent_t start, stop; //used for timings
      HANDLE_ERROR( cudaEventCreate(&start) );
      HANDLE_ERROR( cudaEventCreate(&stop) );
      HANDLE_ERROR( cudaEventRecord(start, 0) ); //insert the start event into stream 0


      //start kernel
      dim3 dimGrid(NUM_BLOCKS, 1, 1);
      dim3 dimBlock(BLOCKSIZE, 1, 1);
#ifdef BAD
      sdot_bad<<<dimGrid, dimBlock>>>(num_elements, x_d, y_d, buffer_d);
#else
      sdot_good<<<dimGrid, dimBlock>>>(num_elements, x_d, y_d, buffer_d);
#endif
      HANDLE_ERROR(cudaGetLastError());

      HANDLE_ERROR( cudaEventRecord(stop, 0) ); //insert stop event into stream 0
      HANDLE_ERROR( cudaEventSynchronize(stop) ); //wait till the stop event has occured
      HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); //take time difference between start and stop event

      if(time < min_time)
         min_time = time;

   }
   printf("Time for sdot kernel:  %3.1f ms \n", min_time);

   //copy data from device to host
   HANDLE_ERROR( cudaMemcpy(buffer_h, buffer_d, sizeof(float) * NUM_BLOCKS, cudaMemcpyDeviceToHost) ); 

   //sum over the remaining elements
   float result = 0.0f;
   for(i=0;i < NUM_BLOCKS; ++i)
      result += buffer_h[i];





   //verify the result
   check_result(num_elements, x_h, y_h, result);

      
   //free memory
   free(x_h); free(y_h); free(buffer_h);
   cudaFree(x_d); cudaFree(y_d); cudaFree(buffer_d);

   return 0;
}


float sdot_ref(int num_elements, float * const  x, float* const y)
{
   int i;
   float sum = 0.0f;
   for(i=0;i < num_elements; ++i)
      sum += y[i] * x[i];

   return sum;
}


/// compares the cpu and gpu results element-wise and checks if the maximal relative error is larger than some threshold
void check_result(int num_elements, float * const x, float * const y, float result)
{
   //start cpu calculation
   float cpu_result = sdot_ref(num_elements, x, y);

   if(fabs((result - cpu_result)/cpu_result) < 1e-3)
      printf_green("Result OKAY!\n");
   else
      printf_red("ERROR: Result-check FAILED!\n");
}


