/**
 * Naive SGEMM implementation with CUDA
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 32


void check_result(float alpha, float * const  A, float * const B, float beta, float *
      const C, float * C_sol, int n);


//sgemm for nxn matrix A,B,C
void sgemm_ref(float alpha, float *A, float *B, float beta, float* C, int n);







//sgemm for nxn matrix A,B,C
__global__ void sgemm(float alpha, float *A, float *B, float beta, float* C, int n)
{

   int idx = threadIdx.x + blockDim.x * blockIdx.x;
   int idy = threadIdx.y + blockDim.y * blockIdx.y;

   float sum = 0.0f;
   for(int k=0;k < n; ++k)
      sum += A[idy * n + k] * B[k * n + idx];

   C[idy * n + idx] = alpha * sum + beta * C[idy * n + idx];

}
















int main(int argc, char** argv)
{
   int i;
   int n = 1024;
   if(argc > 1)
      n = atoi(argv[1]);

   if(n%BLOCKSIZE != 0){
      printf("Matrix size must be divisible by %d\n",BLOCKSIZE);
      exit(-1);
   }

   int size = sizeof(float) * n * n;
   printf("Allocates %f MB\n", 2.0f * ((float)size)/1024/1024);


   //allocate host memory
   float *A_h, *B_h, *C_h, *C_copy;
   A_h = (float*) malloc(size);
   B_h = (float*) malloc(size);
   C_h = (float*) malloc(size);
   C_copy = (float*) malloc(size);


   //init data with random numbers in [0,1]
   for(i=0;i < n*n; ++i){
      A_h[i] = ((float)rand())/RAND_MAX;
      B_h[i] = ((float)rand())/RAND_MAX;
      C_h[i] = ((float)rand())/RAND_MAX;
      C_copy[i] = C_h[i];
   }
   float alpha = ((float)rand())/RAND_MAX; 
   float beta = ((float)rand())/RAND_MAX; 


   //allocate device memory
   float *A_d, *B_d, *C_d;
   HANDLE_ERROR( cudaMalloc(&A_d, size) );
   HANDLE_ERROR( cudaMalloc(&B_d, size) );
   HANDLE_ERROR( cudaMalloc(&C_d, size) );


   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(A_d, A_h, size, cudaMemcpyHostToDevice) ); 
   HANDLE_ERROR( cudaMemcpy(B_d, B_h, size, cudaMemcpyHostToDevice) ); 
   HANDLE_ERROR( cudaMemcpy(C_d, C_h, size, cudaMemcpyHostToDevice) ); 


   float time;
   cudaEvent_t start, stop; //used for timings
   HANDLE_ERROR( cudaEventCreate(&start) );
   HANDLE_ERROR( cudaEventCreate(&stop) );
   HANDLE_ERROR( cudaEventRecord(start, 0) ); //insert the start event into stream 0

   //start 2D kernel
   dim3 dimGrid(n/BLOCKSIZE, n/BLOCKSIZE, 1);
   dim3 dimBlock(BLOCKSIZE, BLOCKSIZE, 1);
   sgemm<<<dimGrid, dimBlock>>>(alpha, A_d, B_d, beta, C_d, n); //this kernel is implicitly inserted into stream 0
   HANDLE_ERROR(cudaGetLastError());

   HANDLE_ERROR( cudaEventRecord(stop, 0) ); //insert stop event into stream 0
   HANDLE_ERROR( cudaEventSynchronize(stop) ); //wait till the stop event has occurred
   HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); //take time difference between start and stop event
   printf("Time:  %3.1f ms \n", time);


   //copy data from device to host
   HANDLE_ERROR( cudaMemcpy(C_h, C_d, size, cudaMemcpyDeviceToHost) ); 


   //verify the result
   check_result(alpha, A_h, B_h, beta, C_copy, C_h, n);

      
   //free memory
   free(A_h); free(B_h); free(C_h); free(C_copy); 
   cudaFree(A_d); cudaFree(B_d); cudaFree(C_d);

   return 0;
}


void sgemm_ref(float alpha, float *A, float *B, float beta, float* C, int n)
{
   int i,j,k;
   for(i=0;i < n; ++i){
      for(j=0;j < n; ++j){
         float sum = 0.0f;
         for(k=0;k < n; ++k)
            sum += A[i * n + k] * B[k * n + j];
         C[i*n + j] = alpha * sum + beta * C[i*n + j];
      }
   }
}

/// C_sol == (alpha * A * B + beta * C) ? 
void check_result(float alpha, float * const  A, float * const B, float beta, float *
      const C, float * C_sol, int n)
{
   //start reference computation
   sgemm_ref(alpha, A, B, beta, C, n);

   int i,j;
   for(i=0;i < n; ++i){
      for(j=0;j < n; ++j){
         if(fabs((C[i * n + j] - C_sol[i * n + j]) / C_sol[i * n + j]) > 1e-5){
            printf_red("ERROR: Result-check FAILED!\n");
            exit(-1);
         }
      }
   }

   printf_green("Result OKAY!\n");
}

