/**
 * Illustrates global memory accesses
 *
 * Lecture notes:
 * 1) Understand the global memory access pattern of transpose_naive. Can we improve it? 
 * 2) Profile the naive version. Have a look at "Global store efficiency"
 * 3) Profile sm version. Have a look at "Shared Memory Replay Overhead", which is an indicator for bank conflicts.
 * 4) Remove bank conflict
 */

#include <cuda_runtime.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "../misc/cuda_util.h"

#define BLOCKSIZE 32


void check_result(float * const  A, float * const B, int n);





/// transposes a n x n matrix
__global__ void transpose_naive(float * const in, float *out, int n);

__global__ void transpose_sm(float * const in, float *out, int n);

__global__ void transpose_sm_impr(float * const in, float *out, int n);


int main(int argc, char** argv)
{
   int i;
   int n= 2048;
   if(argc > 1)
      n = atoi(argv[1]);

   if(n%BLOCKSIZE != 0){
      printf("Matrix size must be divisible by %d\n",BLOCKSIZE);
      exit(-1);
   }

   int size = sizeof(float) * n * n;
   printf("Allocates %f MB\n", 2.0f * ((float)size)/1024/1024);


   //allocate host memory
   float *A_h, *B_h;
   A_h = (float*) malloc(size);
   B_h = (float*) malloc(size);


   //init x and y with random numbers in [0,1]
   for(i=0;i < n*n; ++i){
      A_h[i] = ((float)rand())/RAND_MAX;
   }


   //allocate device memory
   float *A_d, *B_d;
   HANDLE_ERROR( cudaMalloc(&A_d, size) );
   HANDLE_ERROR( cudaMalloc(&B_d, size) );


   //copy data from host to device
   HANDLE_ERROR( cudaMemcpy(A_d, A_h, size, cudaMemcpyHostToDevice) ); 


   float min_time = FLT_MAX;
   int nIter = 20;
   //take minimum over nIter runs
   for(int i=0;i < nIter; ++i){

      float time;
      cudaEvent_t start, stop; //used for timings
      HANDLE_ERROR( cudaEventCreate(&start) );
      HANDLE_ERROR( cudaEventCreate(&stop) );
      HANDLE_ERROR( cudaEventRecord(start, 0) ); //insert the start event into stream 0

      //start 2D kernel
      dim3 dimGrid(n/BLOCKSIZE, n/BLOCKSIZE, 1);
      dim3 dimBlock(BLOCKSIZE, BLOCKSIZE, 1);

//      transpose_naive<<<dimGrid, dimBlock>>>(A_d, B_d, n); 
      transpose_sm<<<dimGrid, dimBlock>>>(A_d, B_d, n); 
      HANDLE_ERROR(cudaGetLastError());

      HANDLE_ERROR( cudaEventRecord(stop, 0) ); //insert stop event into stream 0
      HANDLE_ERROR( cudaEventSynchronize(stop) ); //wait till the stop event has occured
      HANDLE_ERROR( cudaEventElapsedTime(&time, start, stop) ); //take time difference between start and stop event

      if(time < min_time)
         min_time = time;
   }
   printf("Time:  %3.1f ms \n", min_time);


   //copy data from device to host
   HANDLE_ERROR( cudaMemcpy(B_h, B_d, size, cudaMemcpyDeviceToHost) ); 


   //verify the result
   check_result(A_h, B_h, n);

      
   //free memory
   free(A_h); 
   cudaFree(A_d); cudaFree(B_d);

   return 0;
}




/// Is A really the transpose of B?
void check_result(float * const  A, float * const B, int n)
{
   int i,j;
   for(i=0;i < n; ++i)
      for(j=0;j < n; ++j)
         if(fabs((A[i * n + j] - B[j * n + i])/B[j * n + i]) > 1e-5){
            printf_red("ERROR: Result-check FAILED!\n");
            exit(-1);
         }

   printf_green("Result OKAY!\n");
}







/// transposes a n x n matrix
__global__ void transpose_naive(float * const in, float *out, int n)
{

   int idx = blockIdx.x * blockDim.x + threadIdx.x; 
   int idy = blockIdx.y * blockDim.y + threadIdx.y;

   //consecutive threads
   int index_in = n * idy + idx; 

   //non-consecutive threads
   int index_out = n * idx + idy;

   out[index_out] = in[index_in];

}






















__global__ void transpose_sm(float * const in, float *out, int n)
{

   int idx = blockIdx.x * BLOCKSIZE + threadIdx.x; 
   int idy = blockIdx.y * BLOCKSIZE + threadIdx.y;

   //consecutive threads
   int index = n * idy + idx; 

   __shared__ float tile[BLOCKSIZE][BLOCKSIZE];

   // load tile to shared memory (already transposed)
   tile[threadIdx.x][threadIdx.y] = in[index];

   //make sure that all the data is loaded to the tile
   __syncthreads();


   //compute transposed 
   idx = blockIdx.y * BLOCKSIZE + threadIdx.x; 
   idy = blockIdx.x * BLOCKSIZE + threadIdx.y;

   //consecutive threads
   index = n * idy + idx; 

   out[index] = tile[threadIdx.y][threadIdx.x];

}













__global__ void transpose_sm_impr(float * const in, float *out, int n)
{

   int idx = blockIdx.x * BLOCKSIZE + threadIdx.x; 
   int idy = blockIdx.y * BLOCKSIZE + threadIdx.y;

   //consecutive threads
   int index = n * idy + idx; 

   __shared__ float tile[BLOCKSIZE][BLOCKSIZE+1];

   // load tile to shared memory (already transposed)
   tile[threadIdx.x][threadIdx.y] = in[index];

   //make sure that all the data is loaded to the tile
   __syncthreads();


   //compute transposed 
   idx = blockIdx.y * BLOCKSIZE + threadIdx.x; 
   idy = blockIdx.x * BLOCKSIZE + threadIdx.y;

   //consecutive threads
   index = n * idy + idx; 

   out[index] = tile[threadIdx.y][threadIdx.x];

}



